@@ -19,17 +19,21 @@ import mockmip
 import pyutilib.services
 import pyutilib.common
 import pyutilib.misc
-import pyutilib.plugin.core
+import pyutilib.component.core
 import string
 import re
 
 import xml.dom.minidom
 
+import time
+
 
 class CPLEX(ILMLicensedSystemCallSolver):
     """The CPLEX LP/MIP solver
     """
 
+    pyutilib.component.core.alias('cplex', doc='Shell interface to the CPLEX LP/MIP solver')
+
     def __init__(self, **kwds):
         #
         # Call base class constructor
@@ -57,6 +61,14 @@ class CPLEX(ILMLicensedSystemCallSolver):
         self._valid_result_formats[ProblemFormat.cpxlp] = [ResultsFormat.soln]
         self._valid_result_formats[ProblemFormat.mps] = [ResultsFormat.soln]
 
+        # Note: Undefined capabilities default to 'None'
+        self._capabilities = pyutilib.misc.Options()
+        self._capabilities.linear = True
+        self._capabilities.quadratic = True
+        self._capabilities.integer = True
+        self._capabilities.sos1 = True
+        self._capabilities.sos2 = True
+
     #
     # ultimately, this utility should go elsewhere - perhaps on the PyomoModel itself.
     # in the mean time, it is staying here.
@@ -83,7 +95,7 @@ class CPLEX(ILMLicensedSystemCallSolver):
     #
     # write a warm-start file in the CPLEX MST format.
     #
-    def write_warmstart_file(self, instance):
+    def warm_start(self, instance):
 
        import coopr.pyomo.base.var
 
@@ -152,8 +164,12 @@ class CPLEX(ILMLicensedSystemCallSolver):
              if len(args) != 1:
                 raise ValueError, "CPLEX _presolve method can only handle a single problem instance - "+str(len(args))+" were supplied"                  
 
-             if self._hasIntegerVariables(args[0]) is True:          
-                self.write_warmstart_file(args[0])
+             if self._hasIntegerVariables(args[0]) is True:
+                start_time = time.time()
+                self.warm_start(args[0])
+                end_time = time.time()
+                if self._report_timing is True:
+                   print "Warm start write time="+str(end_time-start_time)+" seconds"
           
        # let the base class handle any remaining keywords/actions.
        ILMLicensedSystemCallSolver._presolve(self, *args, **kwds)
@@ -161,7 +177,7 @@ class CPLEX(ILMLicensedSystemCallSolver):
     def executable(self):
         executable = pyutilib.services.registered_executable("cplex")
         if executable is None:
-            pyutilib.plugin.core.PluginGlobals.env().log.error("Could not locate the 'cplex' executable, which is required for solver %s" % self.name)
+            pyutilib.component.core.PluginGlobals.env().log.error("Could not locate the 'cplex' executable, which is required for solver %s" % self.name)
             self.enable = False
             return None
         return executable.get_path()
@@ -258,13 +274,16 @@ class CPLEX(ILMLicensedSystemCallSolver):
         for line in output.split("\n"):
             tokens = re.split('[ \t]+',line.strip())
             if len(tokens) > 3 and tokens[0] == "CPLEX" and tokens[1] == "Error":
+                # IMPT: See below - cplex can generate an error line and then terminate fine, e.g., in CPLEX 12.1.
+                #       To handle these cases, we should be specifying some kind of termination criterion always
+                #       in the course of parsing a log file (we aren't doing so currently - just in some conditions).
                 results.solver.status=SolverStatus.error
                 results.solver.error = " ".join(tokens)
             elif len(tokens) >= 3 and tokens[0] == "ILOG" and tokens[1] == "CPLEX":
                 cplex_version = tokens[2].rstrip(',')
             elif len(tokens) >= 3 and tokens[0] == "Variables":
-                if results.problem.number_of_variables is None: # CPLEX 11.2 and subsequent has two Variables sections.
-                    results.problem.number_of_variables = tokens[2]
+                if results.problem.number_of_variables is None: # CPLEX 11.2 and subsequent versions have two Variables sections in the log file output.
+                    results.problem.number_of_variables = eval(tokens[2])
             # In CPLEX 11 (and presumably before), there was only a single line output to
             # indicate the constriant count, e.g., "Linear constraints : 16 [Less: 7, Greater: 6, Equal: 3]".
             # In CPLEX 11.2 (or somewhere in between 11 and 11.2 - I haven't bothered to track it down
@@ -272,10 +291,10 @@ class CPLEX(ILMLicensedSystemCallSolver):
             # block - which we don't care about. In this case, the line looks like: "Linear constraints :" and
             # that's all.
             elif len(tokens) >= 4 and tokens[0] == "Linear" and tokens[1] == "constraints":
-                results.problem.number_of_constraints = tokens[3]
+                results.problem.number_of_constraints = eval(tokens[3])
             elif len(tokens) >= 3 and tokens[0] == "Nonzeros":
                 if results.problem.number_of_nonzeros is None: # CPLEX 11.2 and subsequent has two Nonzeros sections.                
-                    results.problem.number_of_nonzeros = tokens[2]
+                    results.problem.number_of_nonzeros = eval(tokens[2])
             elif len(tokens) >= 5 and tokens[4] == "MINIMIZE":
                 results.problem.sense = ProblemSense.minimize
             elif len(tokens) >= 5 and tokens[4] == "MAXIMIZE":
@@ -284,10 +303,82 @@ class CPLEX(ILMLicensedSystemCallSolver):
                # technically, I'm not sure if this is CPLEX user time or user+system - CPLEX doesn't appear
                # to differentiate, and I'm not sure we can always provide a break-down.
                results.solver.user_time = eval(tokens[3])
+            elif len(tokens) >= 4 and tokens[0] == "Dual" and tokens[1] == "simplex" and tokens[3] == "Optimal:":
+                results.solver.termination_condition = TerminationCondition.optimal
+                results.solver.termination_message = ' '.join(tokens)
+            elif len(tokens) >= 4 and tokens[0] == "MIP" and tokens[2] == "Integer" and tokens[3] == "infeasible.":
+                # if CPLEX has previously printed an error message, reduce it to a warning -
+                # there is a strong indication it recovered, but we can't be sure.
+                if results.solver.status == SolverStatus.error:
+                   results.solver.status = SolverStatus.warning
+                else:
+                   results.solver.status = SolverStatus.ok
+                results.solver.termination_condition = TerminationCondition.infeasible
+                results.solver.termination_message = ' '.join(tokens)
+            # for the case below, CPLEX sometimes reports "true" optimal (the first case)
+            # and other times within-tolerance optimal (the second case).
+            elif (len(tokens) >= 4 and tokens[0] == "MIP" and tokens[2] == "Integer" and tokens[3] == "optimal") or \
+                 (len(tokens) >= 4 and tokens[0] == "MIP" and tokens[2] == "Integer" and tokens[3] == "optimal,"):
+                # if CPLEX has previously printed an error message, reduce it to a warning -
+                # there is a strong indication it recovered, but we can't be sure.
+                if results.solver.status == SolverStatus.error:
+                   results.solver.status = SolverStatus.warning
+                else:
+                   results.solver.status = SolverStatus.ok                
+                results.solver.termination_condition = TerminationCondition.optimal
+                results.solver.termination_message = ' '.join(tokens)                
+            elif len(tokens) >= 3 and tokens[0] == "Presolve" and tokens[2] == "Infeasible.":
+                # if CPLEX has previously printed an error message, reduce it to a warning -
+                # there is a strong indication it recovered, but we can't be sure.
+                if results.solver.status == SolverStatus.error:
+                   results.solver.status = SolverStatus.warning
+                else:
+                   results.solver.status = SolverStatus.ok                
+                results.solver.termination_condition = TerminationCondition.infeasible
+                results.solver.termination_message = ' '.join(tokens)
+            elif len(tokens) >= 5 and tokens[0] == "Presolve" and tokens[2] == "Unbounded" and tokens[4] == "infeasible.":
+                # if CPLEX has previously printed an error message, reduce it to a warning -
+                # there is a strong indication it recovered, but we can't be sure.
+                if results.solver.status == SolverStatus.error:
+                   results.solver.status = SolverStatus.warning
+                else:
+                   results.solver.status = SolverStatus.ok                
+                # It isn't clear whether we can determine if the problem is unbounded from
+                # CPLEX's output.
+                results.solver.termination_condition = TerminationCondition.unbounded
+                results.solver.termination_message = ' '.join(tokens)
+
+        try:
+            results.solver.termination_message = pyutilib.misc.yaml_fix(results.solver.termination_message)
+        except:
+            pass
         return results
 
     def process_soln_file(self,results):
-        lp_solution=False
+
+        # the only suffixes that we extract from CPLEX are
+        # constraint duals, constraint slacks, and variable 
+        # reduced-costs. scan through the solver suffix list 
+        # and throw an exception if the user has specified 
+        # any others.
+        extract_duals = False
+        extract_slacks = False
+        extract_reduced_costs = False
+        for suffix in self.suffixes:
+           flag=False
+           if re.match(suffix,"dual"):
+                extract_duals = True
+                flag=True
+           if re.match(suffix,"slack"):
+                extract_slacks = True              
+                flag=True
+           if re.match(suffix,"rc"):
+                extract_reduced_costs = True
+                flag=True
+           if not flag:
+              raise RuntimeError,"***CPLEX solver plugin cannot extract solution suffix="+suffix
+
+        lp_solution = False
         if not os.path.exists(self.soln_file):
            return
 
@@ -298,7 +389,6 @@ class CPLEX(ILMLicensedSystemCallSolver):
         mip_problem=False
         for line in INPUT:
             line = line.strip()
-            #print "LLINE",line
             line = line.lstrip('<?/')
             line = line.rstrip('/>?')
             tokens=line.split(' ')
@@ -315,9 +405,9 @@ class CPLEX(ILMLicensedSystemCallSolver):
                       variable_name = field_value
                    elif field_name == "value":
                       variable_value = field_value
-                   elif field_name == "reducedCost":
+                   elif (extract_reduced_costs is True) and (field_name == "reducedCost"):
                       variable_reduced_cost = field_value
-                   elif field_name == "status":
+                   elif (extract_reduced_costs is True) and (field_name == "status"):
                       variable_status = field_value
 
                 # skip the "constant-one" variable, used to capture/retain objective offsets in the CPLEX LP format.
@@ -328,7 +418,7 @@ class CPLEX(ILMLicensedSystemCallSolver):
                        variable.value = eval(variable_value)
                    except:
                        variable.value = variable_value
-                   if not variable_reduced_cost is None:
+                   if (variable_reduced_cost is not None) and (extract_reduced_costs is True):
                         try:
                             variable.rc = eval(variable_reduced_cost)
                             if variable_status is not None:
@@ -342,7 +432,7 @@ class CPLEX(ILMLicensedSystemCallSolver):
                                   variable.urc = 0.0
                         except:
                             raise ValueError, "Unexpected reduced-cost value="+str(variable_reduced_cost)+" encountered for variable="+variable_name
-            elif tokens[0] == "constraint":
+            elif (tokens[0] == "constraint") and ((extract_duals is True) or (extract_slacks is True)):
                 constraint_name = None
                 constraint_dual = None
                 constaint = None # cache the solution constraint reference, as the getattr is expensive.
@@ -352,11 +442,11 @@ class CPLEX(ILMLicensedSystemCallSolver):
                    if field_name == "name":
                       constraint_name = field_value
                       constraint = soln.constraint[constraint_name]
-                   elif field_name == "dual": # for LPs
+                   elif (extract_duals is True) and (field_name == "dual"): # for LPs
                       # assumes the name field is first.
                       if eval(field_value) != 0.0:
                         constraint.dual = eval(field_value)
-                   elif field_name == "slack": # for MIPs
+                   elif (extract_slacks is True) and (field_name == "slack"): # for MIPs
                       # assumes the name field is first.
                       if eval(field_value) != 0.0:
                         constraint.slack = eval(field_value)
@@ -387,10 +477,10 @@ class CPLEX(ILMLicensedSystemCallSolver):
                 
             elif tokens[0].startswith("objectiveValue"):
                 objective_value = (string.strip(tokens[0].split('=')[1])).lstrip("\"").rstrip("\"")
-                soln.objective['f'].value = objective_value
+                soln.objective['f'].value = eval(objective_value)
             elif tokens[0].startswith("solutionStatusString"):
                 solution_status = (string.strip(" ".join(tokens).split('=')[1])).lstrip("\"").rstrip("\"")
-                if solution_status in ["optimal", "integer optimal solution"]:
+                if solution_status in ["optimal", "integer optimal solution", "integer optimal, tolerance"]:
                     soln.status = SolutionStatus.optimal
                     soln.gap = 0.0
                     if results.problem.sense == ProblemSense.minimize:
@@ -402,9 +492,12 @@ class CPLEX(ILMLicensedSystemCallSolver):
                         if "lower_bound" in dir(results.problem):
                             del results.problem.lower_bound
                     mip_problem=True
+                elif solution_status in ["infeasible"]:
+                    soln.status = SolutionStatus.infeasible
+                    soln.gap = None
             elif tokens[0].startswith("MIPNodes"):
                 if mip_problem:
-                    n = eval(string.strip(" ".join(tokens).split('=')[1])).lstrip("\"").rstrip("\"")
+                    n = eval(eval(string.strip(" ".join(tokens).split('=')[1])).lstrip("\"").rstrip("\""))
                     results.solver.statistics.branch_and_bound.number_of_created_subproblems=n
                     results.solver.statistics.branch_and_bound.number_of_bounded_subproblems=n
 
@@ -443,6 +536,8 @@ class MockCPLEX(CPLEX,mockmip.MockMIP):
     """A Mock CPLEX solver used for testing
     """
 
+    pyutilib.component.core.alias('_mock_cplex')
+
     def __init__(self, **kwds):
         try:
            CPLEX.__init__(self, **kwds)
@@ -466,5 +561,4 @@ class MockCPLEX(CPLEX,mockmip.MockMIP):
 
 
 pyutilib.services.register_executable(name="cplex")
-SolverRegistration("cplex", CPLEX)
-SolverRegistration("_mock_cplex", MockCPLEX)
+
