@@ -29,8 +29,11 @@ from coopr.opt.parallel.manager import *
 from coopr.opt.parallel.solver import *
 from coopr.opt.results import SolverResults
 
+
 class SolverManager_Pyro(AsynchronousSolverManager):
 
+    pyutilib.component.core.alias('pyro')
+
     def clear(self):
         """
         Clear manager state
@@ -117,6 +120,6 @@ class SolverManager_Pyro(AsynchronousSolverManager):
                   #self.results[ah.id].write()
                   return ah
 
-if using_pyro:
-    SolverManagerRegistration("pyro", SolverManager_Pyro)
+if not using_pyro:
+    SolverManagerFactory.deactivate('pyro')
 
